﻿# Thanks to L'In20Cible for the discovery of CBasePlayer::PlayerRunCommand

# =============================================================================
# >> IMPORTS
# =============================================================================
# Python imports

# EventScripts imports
import es
import cfglib
import langlib
import playerlib
import usermsg
import vecmath

# SPE imports
import spe
from spe import HookAction
from spe import HookType
from spe.tools.player import SPEPlayer


# =============================================================================
# >> ADDONINFORMATION
# =============================================================================
info = es.AddonInfo()
info.name        = 'Push \'n Pull'
info.version     = '1.0.2'
info.author      = 'Ayuto'
info.url         = 'http://addons.eventscripts.com/addons/user/53757'
info.basename    = 'pushnpull'
info.description = 'A player can push and pull other players'


# =============================================================================
# >> CONFIGURATION
# =============================================================================
config = cfglib.AddonCFG(es.getAddonPath('pushnpull') + '/settings.cfg')

config.text('*'* 17)
config.text('  Push Settings')
config.text('*'* 17)
push         = config.cvar('pnp_push', '1', 'Enables/Disables the push function')
pushnumber   = config.cvar('pnp_push_number', '3', 'Number of pushs per player')
pushrange    = config.cvar('pnp_push_range', '100', 'Range for pushing players')
pushlength   = config.cvar('pnp_push_length', '300', 'Length the players are pushed')
pushheight   = config.cvar('pnp_push_height', '275', 'Height the players are pushed')
pushnext     = config.cvar('pnp_push_next', '1.5', 'Time to next push')
pushteams    = config.cvar('pnp_push_teams', '#t,#ct', 'Teams which can push')
pushteamonly = config.cvar('pnp_push_teamonly', '0', 'Only teammates can be pushed')

config.text('\n')

config.text('*'* 17)
config.text('  Pull Settings')
config.text('*'* 17)
pull         = config.cvar('pnp_pull', '1', 'Enables/Disables the pull function')
pullnumber   = config.cvar('pnp_pull_number', '3', 'Number of pull per player')
pullrange    = config.cvar('pnp_pull_range', '450', 'Range for searching for players')
pulllength   = config.cvar('pnp_pull_length', '375', 'Length the players are pulled')
pullheight   = config.cvar('pnp_pull_height', '275', 'Height the players are pulled')
pullnext     = config.cvar('pnp_pull_next', '2', 'Time to next pull')
pullteams    = config.cvar('pnp_pull_teams', '#t,#ct', 'Teams which can pull')
pullteamonly = config.cvar('pnp_pull_teamonly', '0', 'Only teammates can be pulled')

config.write()
config.execute()

es.set('pushnpull_version', info.version)
es.makepublic('pushnpull_version')


# =============================================================================
# >> GLOBAL VARIABLES
# =============================================================================
strings = langlib.Strings(es.getAddonPath('pushnpull' + '/strings.ini'))
prefix  = '#green[#lightgreen%s %s#green]#default '% (info.name, info.version)

IN_FORWARD   = (1 << 3)
IN_BACK      = (1 << 4)
IN_USE       = (1 << 5)

actionvars = {'push': {'enabled': push, 'number': pushnumber, 'range': pushrange, 'length': pushlength, 'height': pushheight, 'next': pushnext, 'teams': pushteams, 'teamonly': pushteamonly, 'factor': -1},
    'pull': {'enabled': pull, 'number': pullnumber, 'range': pullrange, 'length': pulllength, 'height': pullheight, 'next': pullnext, 'teams': pullteams, 'teamonly': pullteamonly, 'factor': 1}}



# =============================================================================
# >> CLASSES
# =============================================================================
class Players(dict):
    def __getitem__(self, userid):
        userid = int(userid)
        if userid in self:
            return super(Players, self).__getitem__(userid)

        instance = self[userid] = Player(userid)
        return instance

    def __delitem__(self, userid):
        userid = int(userid)
        if userid in self:
            super(Players, self).__delitem__(userid)

players = Players()


class Player(object):
    def __init__(self, userid):
        self.userid     = userid
        self.actionvars = {'push': {'number': 0, 'next': 0},
            'pull': {'number': 0, 'next': 0}}

    def getTargetPlayer(self):
        entity = spe.call('FindPickerEntity', spe.getPlayer(self.userid))
        if not entity:
            return None

        index = spe.getEntityIndex(entity)
        if es.entitygetvalue(index, 'classname') != 'player':
            return None

        return es.getuserid(es.gethandlefromindex(index))

    def getTickIn(self, seconds, tick=None):
        if not tick:
            tick = es.getplayerprop(self.userid, 'CCSPlayer.baseclass.localdata.m_nTickBase')
        cmdrate = int(es.getclientvar(self.userid, 'cl_cmdrate'))
        return tick + cmdrate * seconds if seconds > 0 else tick + 1

    def doAction(self, action, player, tick):
        gvars = actionvars[action]
        if not gvars['enabled']:
            return

        pvars = self.actionvars[action]
        if pvars['next'] > tick:
            return

        if pvars['number'] >= gvars['number']:
            return

        team = player.team
        teamname = getTeamname(team)
        if not teamname or teamname not in str(gvars['teams']).replace(' ', '').split(','):
            return

        target = self.getTargetPlayer()
        if not target:
            return

        if gvars['teamonly']:
            if es.getplayerteam(target) != team:
                return

        coord1 = es.getplayerlocation(target)
        origin = player.origin
        coord2 = origin.x, origin.y, origin.z
        if getDistance(coord1, coord2) > gvars['range']:
            return

        pvars['number'] += 1
        pvars['next'] = self.getTickIn(gvars['next'], tick)
        pull(target, coord1, coord2, gvars['length'], gvars['height'], gvars['factor'])
        tokens   = {'name': es.getplayername(target), 'pcount': pvars['number'], 'gcount': gvars['number']}
        language = playerlib.getPlayer(self.userid).lang
        text     = strings(action, tokens, language)
        usermsg.hudhint(self.userid, text)


# =============================================================================
# >> LOAD & UNLOAD
# =============================================================================
def load():
    spe.parseINI('pushnpull/signatures.ini')
    spe.parseTypesINI('pushnpull/types.ini')
    spe.detourFunction('PlayerRunCommand', HookType.Pre, pre_player_run_command)

def unload():
    spe.undetourFunction('PlayerRunCommand', HookType.Pre, pre_player_run_command)


# =============================================================================
# >> GAME EVENTS
# =============================================================================
def round_start(ev):
    players.clear()

def player_disconnect(ev):
    del players[ev['userid']]

def player_activate(ev):
    userid = ev['userid']
    if es.isbot(userid):
        return

    language = playerlib.getPlayer(userid).lang
    if push:
        es.tell(userid, '#multi', prefix + strings('activatePush', None, language))

    if pull:
        es.tell(userid, '#multi', prefix + strings('activatePull', None, language))


# =============================================================================
# >> CALLBACKS
# =============================================================================
def pre_player_run_command(args):
    ucmd = spe.makeObject('CUserCmd', args[0])
    buttons = ucmd.buttons
    if not buttons & IN_USE or (not buttons & IN_BACK and not buttons & IN_FORWARD):
        return HookAction.Continue, 0

    player = SPEPlayer.get_instance_from_pointer(args[2])
    if player.isdead:
        return HookAction.Continue, 0

    userid = player.userid
    if es.isbot(userid):
        return HookAction.Continue, 0

    players[userid].doAction('push' if buttons & IN_FORWARD else 'pull', player, ucmd.tick_count)
    return HookAction.Continue, 0


# =============================================================================
# >> FUNCTIONS
# =============================================================================
def getTeamname(team):
    teams = {2:'#t', 3:'#ct'}
    if team in teams:
        return teams[team]
    return None

def getDistance(coord1, coord2, level=(0, 1, 2)):
    return sum((float(coord1[x]) - float(coord2[x])) ** 2 for x in level) ** 0.5

def pull(userid, coord1, coord2, length, height, factor):
    vec1   = vecmath.Vector(coord1)
    vec2   = vecmath.Vector(coord2)
    vec    = (vec2 - vec1) * factor
    vec    = vec.setlength(length)
    vec[2] = height
    es.setplayerprop(userid, 'CBasePlayer.localdata.m_vecBaseVelocity', vec)